@@ -23,8 +23,8 @@ class User_phonesManager(AuditManager):
             ('main', 'Главная'),  # 1
         ), default=0, db_index=True)
 
-    @staticmethod
-    def getRecord(record):
+    @classmethod
+    def getRecord(cls, record ) :
         res = {
             'id': record.id,
             'editing': record.editing,
