@@ -4,18 +4,26 @@ from typing import List
 from bitfield import Bit, BitHandler
 from django.core.exceptions import EmptyResultSet
 from django.db import connection, transaction
-from django.db.models import Manager, QuerySet, Q, ForeignKey
+from django.db.models import Manager, QuerySet, Q, ForeignKey, BooleanField
+from django.db.models import Model
 from django.forms import model_to_dict
-
 from isc_common import delAttr, setAttr
-from isc_common.bit import TurnBitOn, TurnBitOff
+from isc_common.bit import TurnBitOn, TurnBitOff, IsBitOnBool
+from isc_common.common import ONE_STAGE_ADDING, DELETE_ONE_STAGE, delete_image_refs
+from isc_common.common.functions import get_dict_only_model_field
 from isc_common.fields.related import ForeignKeyProtect
 from isc_common.http.DSRequest import DSRequest
-from isc_common.number import DelProps
+from isc_common.number import TryToInt
 
 logger = logging.getLogger(__name__)
 
 
+def lazy_bulk_fetch(max_obj, max_count, fetch_func, start=0):
+    counter = start
+    while counter < max_count:
+        yield fetch_func()[counter:counter + max_obj]
+        counter += max_obj
+
 class CommonQuerySet(QuerySet):
 
     def soft_restore(self):
@@ -236,7 +244,7 @@ class CommonQuerySet(QuerySet):
         elif textMatchStyle == 'substring':
             return 'icontains'
         elif textMatchStyle == 'startsWith':
-            return 'istartswith'
+            return 'icontains'
         else:
             return 'exact'
             # raise Exception(f'Неизветный textMatchStyle: {textMatchStyle}')
@@ -289,14 +297,30 @@ class CommonQuerySet(QuerySet):
                     else:
                         _operator = f'__{_operator}'
 
-                    if self.isNotOperator(criterion.get('operator')):
-                        res.add(~Q(**{f"{criterion.get('fieldName')}{_operator}": self.getValue(criterion)}), operator)
+                    if criterion.get('operator') == 'and':
+                        fields = list(filter(lambda x: x.name == 'active', self.model._meta.fields))
+                        if len(fields) > 0:
+                            field = fields[0]
+                            if isinstance(field, BooleanField):
+                                if criterion.get('value') is None or criterion.get('value') == 'Все':
+                                    pass
+                                elif criterion.get('value') is True or criterion.get('value') == 'Да':
+                                    res.add(Q(**{f"{criterion.get('fieldName')}{'__exact'}": True}), operator)
+                                elif criterion.get('value') is False or criterion.get('value') == 'Нет':
+                                    res.add(Q(**{f"{criterion.get('fieldName')}{'__exact'}": False}), operator)
+
                     else:
-                        if _operator == '__icontains':
-                            if self.getValue(criterion) != '':
-                                res.add(Q(**{f"{criterion.get('fieldName')}{_operator}": self.getValue(criterion)}), operator)
+                        if self.isNotOperator(criterion.get('operator')):
+                            res.add(~Q(**{f"{criterion.get('fieldName')}{_operator}": self.getValue(criterion)}), operator)
                         else:
-                            res.add(Q(**{f"{criterion.get('fieldName')}{_operator}": self.getValue(criterion)}), operator)
+                            if _operator == '__icontains':
+                                if self.getValue(criterion) != '':
+                                    if isinstance(criterion.get('value'), list):
+                                        res.add(Q(**{f"{criterion.get('fieldName')}__in": self.getValue(criterion)}), operator)
+                                    else:
+                                        res.add(Q(**{f"{criterion.get('fieldName')}{_operator}": self.getValue(criterion)}), operator)
+                            else:
+                                res.add(Q(**{f"{criterion.get('fieldName')}{_operator}": self.getValue(criterion)}), operator)
 
         return res
 
@@ -360,7 +384,7 @@ class CommonQuerySet(QuerySet):
             for key, value in data.items():
                 if isinstance(value, Bit):
                     criteria.add(Q(**{f"{key}": value}), operator)
-                elif isinstance(value, list) :
+                elif isinstance(value, list):
                     criteria.add(Q(**{f"{key}__in": value}), operator)
                 else:
                     if textMatchStyle == "iexact" and isinstance(value, int):
@@ -534,7 +558,7 @@ class CommonManager(Manager):
 
         res = dict()
         for key, value in data.items():
-            if str(key).find('_', 0, 1) == -1:
+            if key.find('_', 0, 1) == -1:
                 setAttr(res, key, value)
         return res
 
@@ -600,37 +624,33 @@ class CommonManager(Manager):
         self.enabledAll = request.enabledAll
         return self.get_queryset().get_range_rows(start=request.startRow, end=request.endRow, function=function, distinct_field_names=distinct_field_names, json=request.json)
 
-    def createFromRequest(self, request, removed=None, propsArr: List = None):
-        request = DSRequest(request=request)
-        data = request.get_data()
-        _data = data.copy()
+    def get_one_stage(self, data):
+        if data.get('type') != ONE_STAGE_ADDING:
+            return None
 
-        if _data.get('parent') is not None and isinstance(_data.get('parent'), int):
-            setAttr(_data, 'parent_id', _data.get('parent'))
-            delAttr(_data, 'parent')
+        if data.get('model') is None:
+            raise Exception('Не задан параметр model')
 
-        delAttr(_data, 'dataSource')
-        delAttr(_data, 'operationType')
-        delAttr(_data, 'textMatchStyle')
-        delAttr(_data, 'form')
-        self._remove_prop(_data, removed)
+        if data.get('exportModel') is None:
+            raise Exception('Не задан параметр exportModel')
 
-        props = self.get_prp(data=_data, propsArr=propsArr)
-        if props > -1:
-            setAttr(_data, 'props', props)
+        eval(f"exec('{data.get('exportModel')}')")
+        record = eval(f"{data.get('model')}.get_first_stage_element()")
+        return record
 
-        res = super().create(**_data)
-        try:
-            full_name = res.full_name
-        except:
-            full_name = None
-        res = model_to_dict(res)
-        if full_name:
-            setAttr(res, 'full_name', full_name)
+    @classmethod
+    def getRecord(cls, record):
+        return dict(
+            id=record.id
+        )
 
-        setAttr(res, 'isFolder', False)
-        data.update(DelProps(res))
-        return data
+    def getRec(self, id, model):
+
+        if model is None:
+            res = self.getRecord(record=super().get(id=id))
+        else:
+            res = self.getRecord(record=model.objects.get(id=id))
+        return res
 
     def _dataIsArray(self, data):
         return isinstance(data, list)
@@ -649,11 +669,11 @@ class CommonManager(Manager):
                     setAttr(res, key, data.get(key))
         return res
 
-    def clone_data(self, data):
+    def clone_data(self, data, model=None):
         _data = dict()
         if isinstance(data, dict):
             for key, val in data.items():
-                if self.get_field(key):
+                if self.get_field(field_name=key, model=model):
                     setAttr(_data, key, val)
         return _data
 
@@ -674,8 +694,21 @@ class CommonManager(Manager):
                     idx += 1
                 return props
             else:
-                return -1
-        return -1
+                return None
+        return None
+
+    def get_expand_prp(self, data, propsArr, props_name='props'):
+        if isinstance(data, dict):
+            props = data.get(props_name, 0)
+            if isinstance(propsArr, list):
+                idx = 0
+                for prop in propsArr:
+                    setAttr(data, prop, IsBitOnBool(props, idx))
+                    idx += 1
+                return props
+            else:
+                return None
+        return None
 
     def check_data_for_multi_select(self, data):
         for k, v in data.items():
@@ -684,16 +717,66 @@ class CommonManager(Manager):
             break
         return data
 
-    def updateFromRequest(self, request, removed=None, function=None, propsArr: List = None):
-        if not isinstance(request, DSRequest):
-            request = DSRequest(request=request)
+    def get_data_from_GET(self, request):
+        data = dict(zip(request.GET.keys(), request.GET.values()))
+        _data = dict()
+        for k, v in data.items():
+            if v == 'null':
+                setAttr(_data, k, None)
+            elif v == 'false':
+                setAttr(_data, k, False)
+            elif v == 'true':
+                setAttr(_data, k, True)
+            else:
+                setAttr(_data, k, TryToInt(v))
+        return _data
+
+    def _createFromRequest(self, request, removed=None, propsArr: List = None):
+
+        request = DSRequest(request=request)
         data = request.get_data()
 
+        one_stage_record = self.get_one_stage(data=data)
+        if one_stage_record is not None:
+            self.get_expand_prp(data=one_stage_record, propsArr=propsArr)
+            return one_stage_record
+
+        _data = data.copy()
+        setAttr(_data, 'editor', request.user)
+
+        if _data.get('parent') is not None and isinstance(_data.get('parent'), int):
+            setAttr(_data, 'parent_id', _data.get('parent'))
+            delAttr(_data, 'parent')
+
+        self._remove_prop(_data, removed)
+
+        props = self.get_prp(data=data, propsArr=propsArr)
+        if props is not None:
+            setAttr(_data, 'props', props)
+
+        record = get_dict_only_model_field(data=_data, model=self.model, exclude=['id'])
+        res = super().create(**record)
+
+        return res
+
+    def createFromRequest(self, request, model=None, removed=None, propsArr: List = None):
+        res = self._createFromRequest(request=request, removed=removed, propsArr=propsArr)
+        res = self.getRec(id=res.id, model=model)
+        return res
+
+    def _updateFromRequest(self, request, removed=None, function=None, propsArr: List = None):
+        _request = DSRequest(request=request)
+        data = _request.get_data()
+        if data is None:
+            data = self.get_data_from_GET(request=request)
+
         data = self.check_data_for_multi_select(data=data)
 
         _data = data.copy()
+        setAttr(_data, 'editor', request.user)
 
         if self._dataIsArray(data):
+            res_arr = []
             data = list(data.values())
             for data_item in data:
                 self._remove_prop(data_item, removed)
@@ -704,38 +787,54 @@ class CommonManager(Manager):
                     else:
                         props = self.get_prp(data=data_item, propsArr=propsArr)
                         cloned_data = self.clone_data(data)
-                        if props > -1:
+                        if props is not None:
                             setAttr(cloned_data, 'props', props)
 
                         if data_item.get('id'):
-                            super().update_or_create(id=data_item.get('id'), defaults=cloned_data)
+                            res, _ = super().update_or_create(
+                                id=data_item.get('id'),
+                                defaults=get_dict_only_model_field(data=cloned_data, model=self.model, exclude=['id'])
+                            )
                         else:
-                            super().create(**cloned_data)
+                            res = super().create(**cloned_data)
+                        res_arr.append(res.id)
+            return res_arr
         else:
             self._remove_prop(data, removed)
             data = self._remove_prop_(data)
             if function:
-                function(data)
+                res = function(data)
             else:
                 props = self.get_prp(data=data, propsArr=propsArr)
                 cloned_data = self.clone_data(data)
 
-                if props > -1:
+                if props is not None:
                     setAttr(cloned_data, 'props', props)
 
-                id = cloned_data.get('id')
-                if id is not None:
+                _id = cloned_data.get('id')
+                if _id is not None:
                     delAttr(cloned_data, 'id')
                     parent = cloned_data.get('parent')
                     if isinstance(parent, int):
                         delAttr(cloned_data, 'parent')
                         setAttr(cloned_data, 'parent_id', parent)
-                    res = super().filter(id=id).update(**cloned_data)
+
+                    record = get_dict_only_model_field(data=cloned_data, model=self.model, exclude=['id'])
+                    res = super().filter(id=_id).update(**record)
+                    res = super().get(id=_id)
                     # print(f'res: {res}')
                 else:
-                    new_item = super().create(**cloned_data)
-                    setAttr(_data, 'id', new_item.id)
-        return _data
+                    res = super().create(**cloned_data)
+            return res
+
+    def updateFromRequest(self, request, model=None, removed=None, function=None, propsArr: List = None):
+        res = self._updateFromRequest(request=request, removed=removed, function=function, propsArr=propsArr)
+        if isinstance(res, list):
+            res = list(map(lambda id: self.getRec(id=id, model=model), res))
+            return res
+        else:
+            res = self.getRec(id=res.id if isinstance(res, Model) else res.get('id'), model=model)
+            return res
 
     def copyFromRequest(self, request):
         request = DSRequest(request=request)
@@ -868,27 +967,80 @@ class CommonManager(Manager):
                         setAttr(_res, key, value)
         return _res
 
-    def deleteFromRequest(self, request, removed=None, ):
+    def delete_image_refs(self, request):
+        data = request.get_data()
+        if data.get('mode') == delete_image_refs:
+            if data.get('id') is None:
+                raise Exception('Не указан параметр id')
+
+            image_id = data.get(f"{data.get('keyImage')}_image_id")
+            if image_id is None:
+                raise Exception('Не указан параметр image_id')
+
+            if data.get('model_images') is None:
+                raise Exception('Не указан параметр model_images')
+
+            if data.get('import_model_images') is None:
+                raise Exception('Не указан параметр import_model_images')
+
+            eval(f"exec('{data.get('import_model_images')}')")
+            res = eval(f"{data.get('model_images')}.objects.filter( main_model_id={data.get('id')}, image_id={image_id}, deleted_at__isnull = False).count()")
+            res1 = eval(f"{data.get('model_images')}.objects.filter( main_model_id={data.get('id')}, image_id={image_id}, deleted_at__isnull = True).count()")
+
+            if res == 0 and res1 == 0:
+                raise Exception(f"Для  main_model_id={data.get('id')} image_id={image_id} не найден")
+
+            if res != 0:
+                res = [eval(f"{data.get('model_images')}.objects.filter( main_model_id={data.get('id')}, image_id={image_id} ).delete()")[0]]
+            else:
+                res = eval(f"{data.get('model_images')}.objects.filter( main_model_id={data.get('id')}, image_id={image_id} ).soft_delete()")
+            if res is not None:
+                return len(res)
+        return None
+
+    def delete_one_staage(self, request):
+        data = request.get_data()
+        if data.get('mode') == DELETE_ONE_STAGE:
+            id = data.get('id')
+            return super().filter(id=id).delete()
+        return None
+
+    def deleteFromRequest(self, request, removed=None, model=None):
         request = DSRequest(request=request)
-        res = 0
         tuple_ids = request.get_olds_tuple_ids()
         with transaction.atomic():
+
+            res = self.delete_image_refs(request=request)
+            if res is not None:
+                return res
+
+            res = self.delete_one_staage(request=request)
+            if res is not None:
+                return res
+            res = 0
+
             for id, mode in tuple_ids:
-                if mode == 'hide':
-                    super().filter(id=id).soft_delete()
-                    res += 1
-                elif mode == 'visible':
-                    super().filter(id=id).soft_restore()
+                if model is not None and issubclass(model, Model):
+                    if mode == 'hide':
+                        model.objects.filter(id=id).soft_delete()
+                        res += 1
+                    elif mode == 'visible':
+                        model.objects.filter(id=id).soft_restore()
+                    else:
+                        qty, _ = model.objects.filter(id=id).delete()
+                        res += qty
                 else:
-                    qty, _ = super().filter(id=id).delete()
-                    res += qty
+                    if mode == 'hide':
+                        super().filter(id=id).soft_delete()
+                        res += 1
+                    elif mode == 'visible':
+                        super().filter(id=id).soft_restore()
+                    else:
+                        qty, _ = super().filter(id=id).delete()
+                        res += qty
         return res
 
     def get_ex(self, *args, **kwargs):
-        """
-        Perform the query and return a single object matching the given
-        keyword arguments.
-        """
         clone = super().get_queryset().filter(*args, **kwargs)
         num = len(clone)
         if num >= 1:
